/**
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

package P2PWNC;

import java.io.*;
import java.lang.*;

/**
 * This class provides Base64 encoding and decoding functionality, according to
 * the RCF 2045 specifications.
 *
 * @author Pantelis Frangoudis
 * @author Elias Efstathiou
 * @version 0.0.1
 */
public class Base64 {

        /**
         * Length of the lines of the base64 encoded text. It defaults to 64 characters.
         */
        protected int nLineLength;

        /**
         * Encoding table. It includes the 64 legal ascii characters.
         */
        static protected byte EncTable[]=
		{
			'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
			'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
			'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
			'w', 'x', 'y', 'z', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '+', '/'
		};

        /**
         * Returns true if c is a character that should be ignored during the decoding
         * process (that is, tab, space, cr, lf)
         *
         * @param c char
         * The character that is to be checked
         * @return boolean
         * Returns true if the character should be ignored. Otherwise, it returns false
         */
        protected boolean ShouldBeIgnored(char c) {
		if ( (c == 0) || (c == 10) || (c == 13) || (c == 32) || (c == 9) ) {
			return true;
		}
		return false;
	}

        /**
         * Reads base64-encoded data, removes blank/carriage return/linefeed characters
         * @param b64data String
         * @param inlen int
         * Length of the input data
         * @return B64Data
         * @return Returns a B64Data object which contains the translated base64 byte array and its length.
         * Base64 bytes are translated to the
         * original byte values (e.g. "A" ---> 0, "B" ---> 1, "0" ---> 52). In case an error
         * occurs (e.g. non-ascii character, character that doesn't belong to the translation table)
         * it returns NULL.
         */
	protected B64Data TranslateB64(String b64data, int inlen) {
		int i = 0;
		int j = -1;
		char b64arr[];

                B64Data toReturn;
                int outlen;

		//String ---> character array
		b64arr = b64data.toCharArray();

		byte translated[] = new byte[inlen];
		for (i=0;i<inlen;i++) {
			translated[i] = '\0';
		}
		for (i=0;i<inlen;i++) {
			if (!ShouldBeIgnored(b64arr[i])) {
				//not a whitespace
				j++;

				//we could use something like a lookup table,etc...
				if (  (b64arr[i] >= 65) && (b64arr[i] <= 90) ) {
					//capital A-Z
					translated[j] = (byte)(b64arr[i] - 65);
				}
				else
				if  ((b64arr[i] >= 97) && (b64arr[i] <= 122)) {
					//lowercase a-z
					translated[j] = (byte)(b64arr[i] - 71);
				}
				else
				if ((b64arr[i] >= 48) && (b64arr[i] <= 57)) {
					//0-9
					translated[j] = (byte)(b64arr[i] + 4);
				}
				else
				if (b64arr[i] == 43) {
					//+
					translated[j] = 62;
				}
				else
				if (b64arr[i] == 47) {
					// /
					translated[j] =63;
				}
				else
				if (b64arr[i] == 61) {
					//padding(=)
					translated[j] = 0;
				}
				else {
					//data error
					return null;
				}
			} //else ignore it...
		}
		outlen = j + 1;
                toReturn = new B64Data(translated, outlen);
		return toReturn;
	}

        /**
         * Encodes a 3-byte block to a 4-byte base64-encoded block. This function is
         * used internally by the EncodeData and EncodeFile functions.
         *
         * @param in
         * Input data block. Is is a block of original data that will be converted
         * to a block of 4 characters.
         * @param out byte[]
         * Output of the encoding operation on the input block. out is always
         * a four character byte array.
         * The four characters which are the result of the encoding operation.
         * @param len int
         * Length of the original block (in). Legal values are 1, 2, 3.
         * If len is less than 3 we must pad with =
         */
	protected void EncodeBlock(byte in[], byte out[], int len) {
		out[0] = EncTable[ (in[0] & 255)>>> 2 ];
		out[1] = EncTable[ ( ((in[0] & 255) & 0x03) << 4) | ( ((in[1] & 255) & 0xf0) >>> 4) ];
		out[2] = (len > 1 ? EncTable[ ( ( (in[1] & 255) & 0x0f) << 2) | (((in[2] & 255) & 0xc0) >>> 6) ] : (byte)'=');
		out[3] = (len > 2 ? EncTable[ (in[2] & 255 ) & 0x3f ] : (byte)'=');
	}

        /**
         * Decodes a 4-byte base64-encoded block to raw bytes. This function is
         * used internally by the EncodeData and EncodeFile functions.
         * @param in byte[]
         * 4-byte input array
         * @param out byte[]
         * Raw byte array (the result of the decoding operation of 4 bytes).
         */
        protected void DecodeBlock(byte in[], byte out[]) {
		out[0] = (byte)(in[0] << 2 | in[1] >> 4);
		out[1] = (byte)(in[1] << 4 | in[2] >> 2);
		out[2] = (byte)(((in[2] << 6) & 0xc0) | in[3]);
	}


        /**
         * Default constructor. It sets nLineLength to 64.
         */
        public Base64() {
		this.nLineLength = 64;
        }

        /**
         * Sets nLineLength to the specified parameter. If the value supplied is
         * not a multiple of 4, then the default value of 64 is used.
         * @param nLineLength int
         * Length of the lines of the base64 encoded data.
         */
        public Base64(int nLineLength) {
		if (nLineLength % 4 != 0) {
			//nLineLength must be multiple of 4, else ---> default = 64
			this.nLineLength = 64;
		}
		this.nLineLength = nLineLength;
	}


        /**
         * Calculates the size of the resulting base64 encoded string.
         * (after the encoding the size is about 4/3 higher)
         * @param nDataLen int
         * Length of a raw byte array that will be base64-encoded
         * @return
         * Returns the calculated length
         */
        public int CalculateB64Size(int nDataLen) {
		int calc = 0;
		switch (nDataLen % 3) {
			case 1:
				calc = (4*(nDataLen+2)/3) + 2*((int)Math.ceil(( 4*(nDataLen+2)/3 )/(double)this.nLineLength));
				return (4*(nDataLen+2)/3) + 2*((int)Math.ceil(( 4*(nDataLen+2)/3 )/(double)this.nLineLength));
			case 2:
				return (4*(nDataLen+1)/3) + 2*((int)Math.ceil(( 4.0*(nDataLen+1)/3.0 )/(double)this.nLineLength));
			case 0:
				return (4*nDataLen/3 + 2*((int)Math.ceil((4*nDataLen/3)/(double)this.nLineLength)));
		}

		return 0;
	}


        /**
         * Calculates the size of the raw byte array that will result from a
         * dacoding operation performed on the given string.
         * @param b64 String
         * A base64 encoded string.
         * @return int
         * Returnes the calculated length
         */
        public int CalculateRawSize(String b64) {
		int i = 0;
		int blen = 0;
		for (i=0; i<b64.length(); i++) {
			if (!this.ShouldBeIgnored(b64.charAt(i))) {
				blen ++;
			}
		}

		if (b64.contains("==")) {
			return blen*3/4 - 2;
		}
		if (b64.contains("=")) {
			return blen*3/4 - 1;
		}
		return blen*3/4;
	}


        /**
         * Base64-encodes a raw byte array of datalen length. Data is output in lines
         * of nLineLength characters.
         * @param raw byte[]
         * Input byte array that is to be encoded.
         * @param datalen int
         * Length of the raw data byte array.
         * @return String
         * The base64 encoding of the specified input byte array.
         */
        String EncodeData(byte raw[], int datalen) {
		String b64data;
		int i = 0;
		int j = 0;
		int b64len = this.CalculateB64Size(datalen);
		char b64[] = new char[b64len];
		for (i=0; i<b64len; i++) {
			b64[i] = '\0';
		}

		byte b64block[] = new byte[4];
		byte rawblock[] = new byte[3];

		int curpos = 0; //current position in b64 byte array (i.e. how many bytes we have appended to b64)
		int charsinline = 0; //how many chars are in the current line
		for (i = 0; i<datalen-3; i+=3) {
			//construct "raw" block
			for (j=0;j<3;j++) {
				rawblock[j] = raw[i+j];
			}
			//Encode block of raw bytes into a 4byte char array ("b64block")
			EncodeBlock(rawblock, b64block, 3);
			//append b64block to b64(string to be returned
			for (j=0;j<4;j++) {
				b64[curpos + j] = (char)b64block[j];
			}

			curpos+=4;
			charsinline += 4;
			if (charsinline == this.nLineLength) {
				//we suppose that nLineLength is multiple of 4
				charsinline = 0;
				b64[curpos] = '\r';
				b64[curpos+1] = '\n';
				curpos+=2; //added crlf
			}
		}

		if (datalen % 3 != 0) {
			//if we have spare characters because datalen is not multiple of 3...
			for (j=0;j<(datalen%3);j++) {
				rawblock[j] = raw[i+j];
			}
			EncodeBlock(rawblock, b64block, datalen % 3);
		}
		else {
			for (j=0;j<3;j++) {
				rawblock[j] = raw[i+j];
			}
			EncodeBlock(rawblock, b64block, 3);
		}
		for (j=0;j<4;j++) {
			b64[curpos + j] = (char)b64block[j];
		}

                //removes trailing \0 characters from char array
                for (i=0;i<b64.length;i++) {
                    if (b64[b64.length - 1 - i] != 0) {
                        break;
                    }
                }
		b64data = new String(b64, 0, b64.length - i);
		return b64data;
	}


	//Decodes the b64 data ---> raw
        /**
         * Decodes the base64 input data.
         * @param b64 String
         * Base64 encoding of raw data.
         * @param raw byte[]
         * This byte array stores the result of the decoding operation.
         * @return int
         * Length o9f the decoded data.
         */
        int DecodeData(String b64, byte raw[]) {
		//Call TranslateB64 to remove whitespace+crlf characters and padding
		//and to convert b64 ---> original byte values
		//and then operate on "translated"
		int rawpos = 0;
		int i = 0;
		int j = 0;
                int len = 0;
		int inlen = b64.length();

		B64Data translated = this.TranslateB64(b64, inlen);
                len = translated.getDatalen();

		byte datablock[] = new byte[4];
		byte rawblock[] = new byte[3];

		if (translated == null) {
			//data error - TranslateB64  returned null...
			return -1;
		}

		for (i=0; i<len - 4; i+=4) {
			for (j=0; j<4; j++) {
				datablock[j] = translated.getData()[i+j];
			}
			DecodeBlock(datablock, rawblock);
			for (j=0;j<3;j++) {
				raw[rawpos + j] = rawblock[j];
			}
			rawpos+=3;
		}

		if (b64.contains("==")) {
			for (j=0; j<4; j++) {
				datablock[j] = translated.getData()[i+j];
			}
			DecodeBlock(datablock, rawblock);
			raw[rawpos] = rawblock[0];
			rawpos++;
		}
		else
		if (b64.contains("=")) {
			for (j=0; j<4; j++) {
				datablock[j] = translated.getData()[i+j];
			}
			DecodeBlock(datablock, rawblock);
			for (j=0;j<2;j++) {
				raw[rawpos + j] = rawblock[j];
			}
			rawpos+=2;
		}
		else {
			for (j=0; j<4; j++) {
				datablock[j] = translated.getData()[i+j];
			}
			DecodeBlock(datablock, rawblock);
			for (j=0;j<3;j++) {
				raw[rawpos + j] = rawblock[j];
			}
			rawpos+=3;
		}
		return rawpos;
	}

        /**
         * Base64-encodes the input file
         * @param infile String
         * Path of the file that is to be encoded.
         * @param outfile String
         * Path of the file where the encoded data will be stored.
         * @return int
         * Length of the encoded data.
         */
        public int EncodeFile(String infile, String outfile) {
		FileInputStream fin;
		FileOutputStream fout;

		int i = 0;
		int j = 0;

		int charsinline = 0; //how many chars are in the current line

		byte rawblock[] = new byte[3];
		byte b64block[] = new byte[4];

		int bytesread = 0;

		try {
		 fin = new FileInputStream(infile);
		}
		catch (FileNotFoundException e) {
			System.err.println("EncodeFile: File not found");
			return -1;
		}

		try {
		 fout = new FileOutputStream(outfile);
		}
		catch (FileNotFoundException e) {
			System.err.println("EncodeFile: Error opening file");
			return -1;
		}

		try {
			while ( (bytesread = fin.read(rawblock)) != -1) {
				//while (!EOF)
				//Encode block of raw bytes into a 4byte char array ("b64block")
				EncodeBlock(rawblock, b64block, bytesread);
				//writes 4-byte block to outfile
				fout.write(b64block);
				charsinline += 4;
				if (charsinline == this.nLineLength) {
					//we suppose that nLineLength is multiple of 4
					charsinline = 0;
					fout.write((byte)'\r');
					fout.write((byte)'\n');
				}
			}
		}
		catch (IOException e) {
			System.err.println("EncodeFile: Error encoding file");
			return -1;
		}

		try {
			fin.close();
		}
		catch (IOException e) {
			System.err.println("EncodeFile: Error closing input file");
			return -1;
		}

		try {
			fout.close();
		}
		catch (IOException e) {
			System.err.println("EncodeFile: Error closing output file");
			return -1;
		}

		return 0;
	}

        /**
         * Decodes a base64 encoded input file.
         * @param infile String
         * The path of the base64 encoded input file.
         * @param outfile String
         * The path of the decoded output file.
         * @return int
         * Length of the decoded file.
         */
        public int DecodeFile(String infile, String outfile) {
		FileInputStream fin;
		FileOutputStream fout;

		int j = 0;
		int blocklen = 0; //current length of current b64 block

		byte rawblock[] = new byte[3];
		byte b64block[] = new byte[4];
		B64Data translated;
		int nextbyte;

		int len[] = new int[1];

		try {
		 fin = new FileInputStream(infile);
		}
		catch (FileNotFoundException e) {
			System.err.println("DecodeFile: File not found");
			return -1;
		}

		try {
		 fout = new FileOutputStream(outfile);
		}
		catch (FileNotFoundException e) {
			System.err.println("DecodeFile: Error opening file");
			return -1;
		}
		try {
			while ( (nextbyte = fin.read()) != -1) {
				if (!this.ShouldBeIgnored((char)nextbyte)) {
					//if next byte isn't whitespace, etc...
					b64block[blocklen] = (byte)nextbyte;
					blocklen++;
				}
				if (blocklen == 4) {
					translated = this.TranslateB64(new String(b64block), 4);
					if (translated == null) {
						//data error - TranslateB64  returned null...
						System.out.println("data error\n");
						return -1;
					}
					DecodeBlock(translated.getData(), rawblock);
					if (new String(b64block).contains("==")) {
						fout.write(rawblock[0]);
					}
					else
					if (new String(b64block).contains("=")) {
						for (j=0;j<2;j++) {
							fout.write(rawblock[j]);
						}
					}
					else {
						for (j=0;j<3;j++) {
							fout.write(rawblock[j]);
						}
					}

					blocklen = 0;
				}
			}
		}
		catch (IOException e) {
			System.err.println("DecodeFile: Error decoding file");
			return -1;
		}


		try {
			fin.close();
		}
		catch (IOException e) {
			System.err.println("DecodeFile: Error closing input file");
			return -1;
		}

		try {
			fout.close();
		}
		catch (IOException e) {
			System.err.println("DecodeFile: Error closing output file");
			return -1;
		}

		return 0;

	}

	public static void main(String args[]) throws IOException {
		Base64 encoder = new Base64();
		if (args.length < 3) {
			System.err.println("Usage: java Base64 -[d/e] <infile> <outfile>");
			return;
		}

		if (args[0].compareTo("-d") == 0) {
			encoder.DecodeFile(args[1], args[2]);
		}
		else
		if (args[0].compareTo("-e") == 0) {
			encoder.EncodeFile(args[1], args[2]);
		}
		else {
			System.err.println("Usage: java -[d/e] <infile> <outfile>");
		}

		return;
	}

}
