/**
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

package P2PWNC;

import java.io.*;
import java.net.*;
import java.util.Date;
import java.security.interfaces.*;
import java.security.*;
import javax.crypto.*;

/**
 * This class provides the functionality of a P2PWNC protocol client program.
 * That is, it provides functions to connect to the service provider and maintain a session.
 *
 * @author Pantelis Frangoudis
 * @author Elias Efstathiou
 * @version 0.0.1
 */

public class P2PWNCClient extends P2PWNCProto {

  /**
   * Service provider public key. This is obtained when the provider first asks
   * the client for a receipt (RREQ message).
   */
  protected P2PWNCKey providerKey;

  /**
   * Consumer (client) certificate. This is normally read from a file.
   */
  protected P2PWNCCert clientCert;

  /**
   * Consumer (client) secret key. This is normally read from a file.
   */
  protected P2PWNCKey clientPrivKey;

  /**
   * Session timestamp. This field indicates when the session started. The timestamp is
   * set by the provider, when it responds (CACK) to a client's connection (CONN) message.
   */
  protected int sessionTimestamp;

  /**
   * Current weight of a provider-consumer session (that is the amount of traffic
   * initiated by/forwarded to the consumer. Each time the client receives an RREQ message
   * this field is updated.
   */
  protected int currentWeight;

  /**
   * Instance of the P2PWNCio class, responsible for i/o operations, such as reading keys and certificates,
   * reading/writing messages from/to the network, etc.
   */
  protected P2PWNCio ioHandler;

  /**
   * Instance of the P2PWNCCrypt class, responsible for cryptograhic operations performed by the client (such
   * as receipt signing).
   */
  protected P2PWNCCrypt cryptHandler;

  /**
   * Instance of the P2PWNCConverter class, responsible for conversions of messages from wire (text) to
   * internal representation. For example, when a receipt is to be transmitted to the service provider, it is
   * converted from a P2PWNCReceipt object to a text message.
   */
  protected P2PWNCConverter converter;

  /**
   * Socket through which the communication to the service provider is performed.
   */
  protected Socket clientSocket;

  /**
   * Internet address of the service provider.
   */
  protected InetAddress serverAddr;

  /**
   * Constructs a P2PWNCClient object, reading the client certificate and secret key from files.
   * It also constructs the client's ioHandler, cryptHandler and converter fields and sets
   * providerKey to null and sessionTimestamp and currentWeight to 0.
   * @param certPath String
   * The path of the client certificate file.
   * @param privKeyPath String
   * The path of the file where the secret key of the client is stored.
   */
  P2PWNCClient(String certPath, String privKeyPath) {
		ioHandler = new P2PWNCio();
		cryptHandler = new P2PWNCCrypt();
		converter = new P2PWNCConverter();

		this.providerKey = null;
		this.clientCert = ioHandler.readCert(certPath);
//                try {
//                    this.clientPrivKey = ioHandler.readKeyBin(privKeyPath, 256);
                  this.clientPrivKey = ioHandler.readKey(privKeyPath);
  //              } catch (IOException ex) {
    //            }
		this.sessionTimestamp = 0;
		this.currentWeight = 0;
	}

  /**
   * Connects to the service provider. This function initiates a TCP connection to the specified
   * P2PWNC service provider module, after having initialized the serverAddr and clientSocket
   * fields. It sends a CONN message to the server, reads the server response in order
   * to set the session timestamp.
   *
   * @param hostname String
   * The IP address of the service provider.
   * @param port int
   * The TCP port the service provider listens for client messages.
   * @return boolean
   * Returns true if the connection to the server has succeeded.
   */
  public boolean connectToP2PWNCServer(String hostname, int port) {
          String connTextMsg = null;
          P2PWNCMessage connAck = null;

          //construct server ip addr from "hostname" string
          try {
                  serverAddr = InetAddress.getByName(hostname);
          }
          catch (UnknownHostException e) {
                  System.err.println("connectToP2PWNCServer: Unknown host");
                  return false;
          }

          try {
                  this.clientSocket = new Socket(serverAddr, port);
          }
          catch (IOException e) {
                  System.err.println("connectToP2PWNCServer: Could not connect to server " + serverAddr.getHostName() + " at port " + port);
                  return false;
          }

          //send CONN message
          connTextMsg = this.converter.constructConnMsg(this.clientCert);
          if (!this.ioHandler.sendTextMsg(this.clientSocket, connTextMsg)) {
                  System.err.println("connectToP2PWNCServer: Error sending CONN message to server");
                  return false;
          }

          System.err.println("Sent Message:\n" + connTextMsg + "\n");

          //read server response
          connAck = this.ioHandler.readMessage(this.clientSocket);
          if (connAck == null) {
                  System.err.println("connectToP2PWNCServer: Null CONN response");
                  return false;
          }
          if (connAck.getMessageType() != P2PWNC_MSG_CACK) {
                  System.err.println("connectToP2PWNCServer: Connection not ACKed");
                  return false;
          }

          System.err.println("Received message\n" + connAck.getText() + "\n");
          //received CACK msg
          //now extract timestamp text message
          this.sessionTimestamp = this.converter.parseCACKMsg(connAck.getText());
          if (this.sessionTimestamp == -1) {
                  System.err.println("connectToP2PWNCServer: Error extracting session timestamp from CACK message");
                  return false;
          }
          return true;
  }


        /**
         * Main client function. In this method, there is an infinite loop in which the
         * client waits for RREQ messages on clientSocket. Each time it receives such a message
         * the client has to parse it, extract information about the current session weight (and store it
         * in the currentWeight field), check the server's public key (included in the RREQ message) and finally
         * sign a receipt and send it to the service provider.
         */
        public void P2PWNCClientFunc() {
            P2PWNCMessage msg = null;
            P2PWNCRREQMsgInfo rreqInfo = null;
            String receiptTextMsg = null;
            P2PWNCReceipt rcpt = null;

            while (true) {
                msg = this.ioHandler.readMessage(this.clientSocket);
                if (msg == null) {
                    System.err.println("P2PWNCClientFunc: Null RREQ message received");
                    return;
                }

                System.err.println("Received Message: \n" + msg.getText() + "\n");

                switch (msg.getMessageType()) {
                case P2PWNC_MSG_RREQ:
                    rreqInfo = this.converter.parseRREQMsg(msg.getText());
                    if (rreqInfo.getWeight() == -1) {
                        System.err.println("P2PWNCClientFunc: Error parsing RREQ message");
                        return;
                    }
                    if (rreqInfo.getProviderKey() == null) {
                        System.err.println("P2PWNCClientFunc: Null provider key");
                        return;
                    }
                    //else: OK
                    //TODO: weight check, etc...
                    this.currentWeight = rreqInfo.getWeight();
                    this.providerKey = rreqInfo.getProviderKey();

                    rcpt = this.cryptHandler.constructReceipt(this.clientCert, this.providerKey, this.sessionTimestamp, this.currentWeight, this.clientPrivKey, false);
                    if (rcpt == null) {
                        System.err.println("P2PWNCClientFunc: Error constructing internal receipt structure");
                        return;
                    }
                    receiptTextMsg = this.converter.receiptToTextMsg(rcpt);
                    if (receiptTextMsg == null) {
                        System.err.println("P2PWNCClientFunc: Error converting receipt to text message");
                        return;
                    }

                    if (!this.ioHandler.sendTextMsg(this.clientSocket, receiptTextMsg)) {
                        System.err.println("P2PWNCClientFunc: Error sending message to server");
                        return;
                    }
                    System.err.println("Sent Message\n" + receiptTextMsg + "\n");

                    break;

                    case P2PWNC_MSG_TEAR:
                        System.err.println("P2PWNCClientFunc: Received TEAR message from server. Exiting...");
                        return;
                    default:
                        System.err.println("P2PWNCClientFunc: Unknown message type" + msg.getText());
                        return;
                } //endsw
            } //loop
	}

        public static void printUsage() {
          System.err.print("\n\nUsage: java P2PWNCClient [certificate path]");
          System.err.println(" [private key path] [P2PWNC server IP addr.] [server port(default: 9999)]\n\n");
        }

	public static void main(String args[]) throws IOException,
            SignatureException, InvalidKeyException, NoSuchAlgorithmException {

          if (args.length != 4) {
            System.err.println("Wrong number of arguments");
            printUsage();
            System.exit(1);
          }

          P2PWNCClient client = new P2PWNCClient(args[0], args[1]);
          client.connectToP2PWNCServer(args[2], Integer.parseInt(args[3]));
          client.P2PWNCClientFunc();
          return;
	}
}
