/**
* Copyright (C) 2004-2005 Pantelis Frangoudis, Elias Efstathiou
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

package P2PWNC;

import java.io.*;
import java.util.Date;
import java.util.SimpleTimeZone;
import java.text.*;
import java.util.Locale;

/**
 * This class provides conversion functionality from the internal representation to text messages and
 * vice versa.
 * The methods provided have to do with parsing text messages read from the network (that is from the socket
 * the client listens for and sends messages to the service provider).
 * @author Pantelis Frangoudis, Elias Efstathiou
 * @version 0.0.1
 */
class P2PWNCConverter extends P2PWNCProto {

       /**
        * Base64 encoder/decoder. It is used for encoding of raw data that will be placed
        * in text messages (public keys, certificates, digital signatures)
        */
        protected Base64 b64Encoder;

        /**
         * Constructs a new P2PWNCConverter object, initializing its b64encoder field.
         */
        P2PWNCConverter () {
		b64Encoder = new Base64();
	}

        /**
         * Converts a P2PWNCReceipt object to an RCPT text message to be sent to the service provider.
         * An RCPT text message is composed of the following fields:
         * <table border="0" cellspacing="3" cellpadding="0">
         * <tr bgcolor="#ccccff"><th align="left">Field</th><th align="left">Description</th></tr>
         * <tr><td>Message Type</td><td>RCPT</td></tr>
         * <tr bgcolor="#eeeeff"><td>Content-length</td><td>Length of the message in octets</td>
         * <tr><td>Algorithm</td><td>Algorithm identifier and bit length of signature(for example RSA1024)</td></tr>
         * <tr bgcolor="#eeeeff"><td>Timestamp</td><td>Session timestamp (in RFC822 format)</td></tr>
         * <tr><td>Weight</td><td>Current session traffic (in bytes)</td></tr>
         * <tr bgcolor="#eeeeff"><td>Client certificate</td><td>The client certificate data, that is
         * the certificate issuer's public key, user's public key and issuer signature. These fields are base64-encoded
         * and separated by two CRLF sequences.</td></tr>
         * <tr><td>Service provider public key</td><td>This field is the base64 encoded public key of the service provider, ending
         * with two CRLF sequences</td></tr>
         * <tr bgcolor="#eeeeff"><td>Receipt signature</td><td>This is the base64 signature of the receipt, produced
         * by the service consumer (client) using its private key.</td></tr>
         * </table>
         *
         * @param rcpt P2PWNCReceipt
         * The P2PWNCReceipt object that is to be converted to a text RCPT message
         * @return String
         * Returns the RCPT text message that corresponds to rcpt.
         */
        public String receiptToTextMsg(P2PWNCReceipt rcpt) {
		StringBuffer text = new StringBuffer();
		SimpleDateFormat fmt;
                SimpleTimeZone tz = new SimpleTimeZone(0, "UTC");
		StringBuffer strDate = new StringBuffer();
		StringBuffer strWeight = new StringBuffer();
		StringBuffer algorithm = new StringBuffer();
                StringBuffer encData = new StringBuffer();
                byte[] serializedData = null;

		int clen = 0;

		Date d;

		if (rcpt == null) {
			System.err.println("receiptToTextMsg: Null receipt");
			return null;
		}

		text.append("RCPT P2PWNC/1.0\r\n");

		algorithm.append("Algorithm: ");
		switch(rcpt.receiptSignature.algorithm) {
			case P2PWNC_ALGO_ECC:
				algorithm.append("ECC");
				break;
			case P2PWNC_ALGO_RSA: default:
				algorithm.append("RSA");
				break;
		}
		algorithm.append(Integer.toString(rcpt.receiptSignature.bits) + "\r\n");
		clen += algorithm.length();

		//convert timestamp to rfc822 format
		d = new Date( ((long)rcpt.timestamp)*1000 );
		fmt = new SimpleDateFormat("EEE, d MMM yyyy HH:mm:ss Z", Locale.US);
                fmt.setTimeZone(tz);
		strDate.append("Timestamp: ");
		fmt.format(d, strDate, new FieldPosition(0));
		strDate.append("\r\n");
		clen += strDate.length();

		strWeight.append("Weight: " + Integer.toString(rcpt.weight) + "\r\n");
		//Now add the b64-encoded keys, signatures to the message
		clen += strWeight.length();

                //serialize keys, signatures, etc
                //Order: Consumer cert, Provider public key, Receipt Signature
                //first calculate their total length
                int serializedDatalen =
                    rcpt.getConsumerCert().getTeamPublicKey().getDatalen() +
                    rcpt.getConsumerCert().getUserPublicKey().getDatalen() +
                    rcpt.getConsumerCert().getTeamSignature().getDatalen() +
                    rcpt.getProviderPublicKey().getDatalen() +
                    rcpt.getReceiptSignature().getDatalen();

                //allocate space for the serialized data
                serializedData = new byte[serializedDatalen];

                //temp data: the array that is now being appended to the serialized data byte arr.
                //tempdatalen: length of "tempData"
                byte []tempData = rcpt.getConsumerCert().getTeamPublicKey().getData();
                int tempDataLen = rcpt.getConsumerCert().getTeamPublicKey().getDatalen();
                int pos = 0;
                //append consumer team public key
                for (int i=0; i<tempDataLen; i++) {
                  serializedData[pos] = tempData[i];
                  pos++;
                }

                //append consumer user public key
                tempData = rcpt.getConsumerCert().getUserPublicKey().getData();
                tempDataLen = rcpt.getConsumerCert().getUserPublicKey().getDatalen();
                for (int i=0; i<tempDataLen; i++) {
                  serializedData[pos] = tempData[i];
                  pos++;
                }

                //append consumer certificate signature
                tempData = rcpt.getConsumerCert().getTeamSignature().getData();
                tempDataLen = rcpt.getConsumerCert().getTeamSignature().getDatalen();
                for (int i=0; i<tempDataLen; i++) {
                  serializedData[pos] = tempData[i];
                  pos++;
                }

                //done with user cert...now append provider team public key
                tempData = rcpt.getProviderPublicKey().getData();
                tempDataLen = rcpt.getProviderPublicKey().getDatalen();
                for (int i=0; i<tempDataLen; i++) {
                  serializedData[pos] = tempData[i];
                  pos++;
                }

                //finally, add the receipt signature to "serializedData"
                tempData = rcpt.getReceiptSignature().getData();
                tempDataLen = rcpt.getReceiptSignature().getDatalen();
                System.err.println(tempDataLen);
                for (int i=0; i<tempDataLen; i++) {
                  serializedData[pos] = tempData[i];
                  pos++;
                }

                //Now that we've serialized data, we b64-encode them
                encData.append(b64Encoder.EncodeData(serializedData, serializedDatalen));

                clen += encData.length();

		text.append("Content-length: " + Integer.toString(clen) + "\r\n");
		text.append(algorithm);
		text.append(strDate);
		text.append(strWeight);
		text.append(encData);
		return text.toString();
	}

        /**
         * Parses an RCPT text message to the correspondent P2PWNCReceipt object.
         * @param text String
         * The text message to be parsed.
         * @return P2PWNCReceipt
         * The P2PWNCReceipt object which is the result of the parsing operation.
         */
        public P2PWNCReceipt textMsgToReceipt(String text) {
          //HAS ERRORS!!!!! (but its not being used anyway...-- TODO
		P2PWNCReceipt rcpt;

		P2PWNCCert cert;
		P2PWNCKey teampub;
		P2PWNCKey userpub;
		P2PWNCSignature teamsig;

		P2PWNCKey prov;

		P2PWNCSignature rsig;

		int datalen = 0;
		byte data[];

		SimpleDateFormat fmt = new SimpleDateFormat("EEE, d MMM yyyy HH:mm:ss Z");
		Date d;

		int pos = 0;
		int crlfpos = 0;
		int algorithm = 0;
		int bits = 0;
		int timestamp = 0;
		int weight = 0;

		String strAlgorithm;
		//check message type (must be RCPT)
		if (text.substring(0, 17).compareToIgnoreCase("RCPT P2PWNC/1.0\r\n") != 0) {
			System.err.println("Error parsing message type");
			return null;
		}
		//move position to the start of the next message header ("Content-length")
		pos += 17;

		//read content length
		if (text.substring(pos, "Content-length: ".length() + pos).compareToIgnoreCase("Content-length: ") != 0) {
			System.err.println("Could not find Content-length header");
			return null;
		}
		if ( (crlfpos = text.substring(pos).indexOf("\r\n")) == -1) {
			return null;
		}
		pos += crlfpos + 2;

		//read algorithm identifier
		if (text.substring(pos, pos + "Algorithm: ".length()).compareToIgnoreCase("Algorithm: ") != 0) {
			System.err.println("Could not find algorithm header");
			return null;
		}
		pos += "Algorithm: ".length();

		strAlgorithm = text.substring(pos, pos + 3);
		if (strAlgorithm.compareToIgnoreCase("RSA") != 0) {
			if (strAlgorithm.compareToIgnoreCase("ECC") != 0) {
				return null;
			}
			else {
				algorithm = P2PWNC_ALGO_ECC;
			}
		}
		else {
			algorithm = P2PWNC_ALGO_RSA;
		}
		pos += 3;

		//read bits
		if ( (crlfpos = text.substring(pos).indexOf("\r\n")) == -1) {
			return null;
		}
		try {
			bits = Integer.parseInt(text.substring(pos, pos + crlfpos));
		}
		catch (NumberFormatException e) {
			System.err.println("Error parsing bits field");
			return null;
		}
		pos += crlfpos + 2;

		//read timestamp
		if (text.substring(pos, pos + "Timestamp: ".length()).compareToIgnoreCase("Timestamp: ") != 0) {
			return null;
		}
		pos += "Timestamp: ".length();
		if ( (crlfpos = text.substring(pos).indexOf("\r\n")) == -1) {
			return null;
		}
		fmt.setLenient(true);
		if ( (d = fmt.parse(text.substring(pos, pos + crlfpos), new ParsePosition(0))) == null) {
			return null;
		}
		else {
			timestamp = (int)(d.getTime()/1000);
		}
		pos += crlfpos + 2;

		//read weight
		if (text.substring(pos, pos + "Weight: ".length()).compareToIgnoreCase("Weight: ") != 0) {
			return null;
		}
		pos += "Weight: ".length();
		if ( (crlfpos = text.substring(pos).indexOf("\r\n")) == -1) {
			return null;
		}
		try {
			weight = Integer.parseInt(text.substring(pos, pos + crlfpos));
		}
		catch (NumberFormatException e) {
			System.err.println("Error parsing weight field");
			return null;
		}
		pos += crlfpos + 2;

		//now read b64 encoded fields (cert, key, signature)
		crlfpos = text.substring(pos).indexOf("\r\n\r\n");
		if (crlfpos == -1) {
			return null;
		}
		//read teampubkey of certificate
		datalen = b64Encoder.CalculateRawSize(text.substring(pos, pos + crlfpos));
		teampub = new P2PWNCKey(algorithm, bits, datalen);
		data = new byte[datalen];
		b64Encoder.DecodeData(text.substring(pos, pos + crlfpos), data);
		teampub.setData(data);

		//read userpubkey of certificate
		pos += crlfpos + 4;
		crlfpos = text.substring(pos).indexOf("\r\n\r\n");
		if (crlfpos == -1) {
			return null;
		}
		datalen = b64Encoder.CalculateRawSize(text.substring(pos, pos + crlfpos));
		userpub = new P2PWNCKey(algorithm, bits, datalen);
		data = new byte[datalen];
		b64Encoder.DecodeData(text.substring(pos, pos + crlfpos), data);
		userpub.setData(data);

		//read teamsig of certificate
		pos += crlfpos + 4;
		crlfpos = text.substring(pos).indexOf("\r\n\r\n");
		if (crlfpos == -1) {
			return null;
		}
		datalen = b64Encoder.CalculateRawSize(text.substring(pos, pos + crlfpos));
		teamsig = new P2PWNCSignature(algorithm, bits, datalen);
		data = new byte[datalen];
		b64Encoder.DecodeData(text.substring(pos, pos + crlfpos), data);
		teamsig.setData(data);

		//read provider pk
		pos += crlfpos + 4;
		crlfpos = text.substring(pos).indexOf("\r\n\r\n");
		if (crlfpos == -1) {
			return null;
		}
		datalen = b64Encoder.CalculateRawSize(text.substring(pos, pos + crlfpos));
		prov = new P2PWNCKey(algorithm, bits, datalen);
		data = new byte[datalen];
		b64Encoder.DecodeData(text.substring(pos, pos + crlfpos), data);
		prov.setData(data);

		//read receipt signature
		pos += crlfpos + 4;
		datalen = b64Encoder.CalculateRawSize(text.substring(pos));
		rsig = new P2PWNCSignature(algorithm, bits, datalen);
		data = new byte[datalen];
		b64Encoder.DecodeData(text.substring(pos), data);
		rsig.setData(data);

		cert = new P2PWNCCert(teampub, userpub, teamsig);
		rcpt = new P2PWNCReceipt(cert, prov, timestamp, weight, rsig);

		return rcpt;
	}

        /**
         * Constructs a CONN text message. This message includes the certificate of the client
         * that wishes to connect to a service provider. The certificate fields (public keys and
         * signature are base64 encoded and placed in the message separated by a sequence of two
         * CRLF line brakes. A CONN message consists of the following fields:
         * <table border="0" cellspacing="3" cellpadding="0">
         * <tr bgcolor="#ccccff"><th align="left">Field</th><th align="left">Description</th></tr>
         * <tr><td>Message Type</td><td>CONN</td></tr>
         * <tr bgcolor="#eeeeff"><td>Content-length</td><td>Length of the message in octets</td>
         * <tr><td>Client certificate</td><td>The client certificate data, that is
         * the certificate issuer's public key, user's public key and issuer signature. These fields are base64-encoded
         * and separated by two CRLF sequences.</td></tr>
         * </table>
         * @param cert P2PWNCCert
         * The client certificate.
         * @return String
         * The resulting CONN message.
         */
        //TODO: prepei na to ftiaksw na katalavainei thn kainouria morfh mhnymatwn gia na mporei na
        //doulepsei to "update" mode
        public String constructConnMsg(P2PWNCCert cert) {
		StringBuffer text = new StringBuffer();
		StringBuffer algorithm = new StringBuffer();
		StringBuffer encData = new StringBuffer();
                byte []serializedData = null;
                int serializedDatalen = 0;

		int clen = 0;

		text.append("CONN P2PWNC/1.0\r\n");

		algorithm.append("Algorithm: ");
		switch(cert.getTeamSignature().getAlgorithm()) {
			case P2PWNC_ALGO_ECC:
				algorithm.append("ECC");
				break;
			case P2PWNC_ALGO_RSA: default:
				algorithm.append("RSA");
				break;
		}
		algorithm.append(Integer.toString(cert.getTeamSignature().getBits()) + "\r\n");
		clen += algorithm.length();


                //serialize keys, signatures, etc
                //Order: Consumer cert, Provider public key, Receipt Signature
                //first calculate their total length
                serializedDatalen =
                    cert.getTeamPublicKey().getDatalen() +
                    cert.getUserPublicKey().getDatalen() +
                    cert.getTeamSignature().getDatalen();

                //allocate space for the serialized data
                serializedData = new byte[serializedDatalen];

                //temp data: the array that is now being appended to the serialized data byte arr.
                //tempdatalen: length of "tempData"
                byte []tempData = cert.getTeamPublicKey().getData();
                int tempDataLen = cert.getTeamPublicKey().getDatalen();
                int pos = 0;
                //append consumer team public key
                for (int i=0; i<tempDataLen; i++) {
                  serializedData[pos] = tempData[i];
                  pos++;
                }

                //append consumer user public key
                tempData = cert.getUserPublicKey().getData();
                tempDataLen = cert.getUserPublicKey().getDatalen();
                for (int i=0; i<tempDataLen; i++) {
                  serializedData[pos] = tempData[i];
                  pos++;
                }

                //append consumer certificate signature
                tempData = cert.getTeamSignature().getData();
                tempDataLen = cert.getTeamSignature().getDatalen();
                for (int i=0; i<tempDataLen; i++) {
                  serializedData[pos] = tempData[i];
                  pos++;
                }

                //Now that we've serialized data, we b64-encode them
                encData.append(b64Encoder.EncodeData(serializedData, serializedDatalen));

                clen += encData.length();

                text.append("Content-length: " + Integer.toString(clen) + "\r\n");
		text.append(algorithm);
		text.append(encData);

		return text.toString();
	}

        /**
         * Parses a CACK text message. A CACK message is the acknowledgement of a CONN
         * message sent by a client to a service provider. This message
         * contains a timestamp set by the service provider indicating the time beginning of the session.
         * This timestamp will be the same in the subsequent RCPT messages sent by the client
         * to refresh the session. A CACK message contains the following fields:
         * <table border="0" cellspacing="3" cellpadding="0">
         * <tr bgcolor="#ccccff"><th align="left">Field</th><th align="left">Description</th></tr>
         * <tr><td>Message Type</td><td>CACK</td></tr>
         * <tr bgcolor="#eeeeff"><td>Content-length</td><td>Length of the message in octets</td>
         * <tr bgcolor="#eeeeff"><td>Timestamp</td><td>The session timestamp (in RFC822 format).</td></tr>
         * </table>
         * @param text String
         * The CACK text message to be parsed.
         * @return int
         * The timestamp as the number of seconds since the Epoch.
         */
        int parseCACKMsg(String text) {
		//SimpleDateFormat fmt = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss Z");
                SimpleDateFormat fmt = new SimpleDateFormat("EEE', 'dd' 'MMM' 'yy' 'HH:mm:ss' 'Z", Locale.US);
		Date d;
		int timestamp = 0;

		int pos = 0;
		int crlfpos = 0;

		String strAlgorithm;

		//check message type (must be CACK)
		if (text.substring(0, 17).compareToIgnoreCase("CACK P2PWNC/1.0\r\n") != 0) {
			System.err.println("parseCACKMsg: Error parsing message type");
			return -1;
		}
		//move position to the start of the next message header ("Content-length")
		pos += 17;

		//read content length
		if (text.substring(pos, "Content-length: ".length() + pos).compareToIgnoreCase("Content-length: ") != 0) {
			System.err.println("parseCACKMsg: Could not find Content-length header");
			return -1;
		}
		if ( (crlfpos = text.substring(pos).indexOf("\r\n")) == -1) {
                    System.err.println("parseCACKMsg: Error parsing line break after Content-length header");
                    return -1;
		}
		pos += crlfpos + 2;

		//read timestamp
		if (text.substring(pos, pos + "Timestamp: ".length()).compareToIgnoreCase("Timestamp: ") != 0) {
                    System.err.println("parseCACKMsg: Could not find Timestamp header");
                    return -1;
		}
		pos += "Timestamp: ".length();
		fmt.setLenient(true);

//		if ( (d = fmt.parse(text.substring(pos), new ParsePosition(0))) == null) {
                try {
                    if ((d = fmt.parse(text.substring(pos))) == null) {
                        System.err.println("parseCACKMsg: Error parsing timestamp value");
                        return -1;
                    } else {
                        timestamp = (int) (d.getTime() / 1000);
                    }
                }
                catch (ParseException ex) {
                    System.err.println("parseCACKMsg: Error parsing timestamp string");
                }
		return timestamp;
	}

        /**
         * Parses an RREQ text message. An RREQ message is periodically sent to the client by the
         * service provider and is a request that the client sents back a signed receipt acknowledging
         * that she has consumed a specific amount of service ("weight", measured in bytes). Thus, an RREQ message contains
         * the receipt "weight". It also contains the service provider's public key, which will be stored by
         * the client when she first receives an RREQ message. RREQ message fields are shown in the following
         * table:
         * <table border="0" cellspacing="3" cellpadding="0">
         * <tr bgcolor="#ccccff"><th align="left">Field</th><th align="left">Description</th></tr>
         * <tr><td>Message Type</td><td>RREQ</td></tr>
         * <tr bgcolor="#eeeeff"><td>Content-length</td><td>Length of the message in octets</td>
         * <tr><td>Algorithm</td><td>Algorithm identifier and bit length of the provider's public key (for example RSA1024)</td></tr>
         * <tr bgcolor="#eeeeff"><td>Weight</td><td>Current session traffic (in bytes, as measured by the provider)</td></tr>
         * <tr><td>Service provider public key</td><td>This field is the base64 encoded public key of the service provider</td></tr>
         * </table>

         * @param text String
         * The RREQ text message to be parsed.
         * @return P2PWNCRREQMsgInfo
         * The result of the parsing operation. The returned P2PWNCRREQMsgInfo object contains the
         * extracted provider public key and the value of the weight field of the message.
         */
        P2PWNCRREQMsgInfo parseRREQMsg(String text) {
		P2PWNCKey prov;
		int datalen = 0;
		byte data[];

		int pos = 0;
		int crlfpos = 0;
		int algorithm = 0;
		int bits = 0;
		int weight = 0;

		String strAlgorithm;
		//check message type (must be RREQ)
		if (text.substring(0, 17).compareToIgnoreCase("RREQ P2PWNC/1.0\r\n") != 0) {
			System.err.println("parseRREQMsg: Error parsing message type");
			return null;
		}
		//move position to the start of the next message header ("Content-length")
		pos += 17;

		//read content length
		if (text.substring(pos, "Content-length: ".length() + pos).compareToIgnoreCase("Content-length: ") != 0) {
			System.err.println("parseRREQMsg: Could not find Content-length header");
			return null;
		}
		if ( (crlfpos = text.substring(pos).indexOf("\r\n")) == -1) {
                    System.err.println("parseRREQmsg: Error parsing line break after Content-length header");
                    return null;
		}
		pos += crlfpos + 2;

		//read algorithm identifier
		if (text.substring(pos, pos + "Algorithm: ".length()).compareToIgnoreCase("Algorithm: ") != 0) {
			System.err.println("parseRREQMsg: Could not find algorithm header");
			return null;
		}
		pos += "Algorithm: ".length();

		strAlgorithm = text.substring(pos, pos + 3);
		if (strAlgorithm.compareToIgnoreCase("RSA") != 0) {
			if (strAlgorithm.compareToIgnoreCase("ECC") != 0) {
                            System.err.println("parseRREQmsg: Invalid algorithm identifier");
                            return null;
			}
			else {
				algorithm = P2PWNC_ALGO_ECC;
			}
		}
		else {
			algorithm = P2PWNC_ALGO_RSA;
		}
		pos += 3;

		//read bits
		if ( (crlfpos = text.substring(pos).indexOf("\r\n")) == -1) {
                    System.err.println("parseRREQmsg: Error parsing line break after algorithm header");
                    return null;
		}
		try {
			bits = Integer.parseInt(text.substring(pos, pos + crlfpos));
		}
		catch (NumberFormatException e) {
			System.err.println("parseRREQMsg: Error parsing bits field");
			return null;
		}
		pos += crlfpos + 2;

		//read weight
		if (text.substring(pos, pos + "Weight: ".length()).compareToIgnoreCase("Weight: ") != 0) {
                    System.err.println("parseRREQmsg: Error parsing weight header");
                    return null;
		}
		pos += "Weight: ".length();
		if ( (crlfpos = text.substring(pos).indexOf("\r\n")) == -1) {
                    System.err.println("parseRREQmsg: Error parsing line break after weight header");
                    return null;
		}
		try {
			weight = Integer.parseInt(text.substring(pos, pos + crlfpos));
		}
		catch (NumberFormatException e) {
			System.err.println("parseRREQMsg: Error parsing weight field");
			return null;
		}
		pos += crlfpos + 2;

		//now read provider public key
//                System.out.println(text.substring(pos));
		datalen = b64Encoder.CalculateRawSize(text.substring(pos));
		prov = new P2PWNCKey(algorithm, bits, datalen);
		data = new byte[datalen];
//		b64Encoder.DecodeData(text.substring(pos, pos + crlfpos), data);
                b64Encoder.DecodeData(text.substring(pos), data);
		prov.setData(data);

		return new P2PWNCRREQMsgInfo(prov, weight);
	}

 	public static void main (String args[]) throws IOException {
		Date d = new Date();
		P2PWNCKey pk = new P2PWNCKey(P2PWNC_ALGO_RSA, 1024, "public key".length());
		pk.setData("public key".getBytes());
		P2PWNCKey prkey = new P2PWNCKey(P2PWNC_ALGO_RSA, 1024, "private key".length());
		prkey.setData("private key".getBytes());
		P2PWNCSignature sig = new P2PWNCSignature(P2PWNC_ALGO_RSA, 1024, "cert sig".length());
		sig.setData("cert sig".getBytes());
		P2PWNCCert cert = new P2PWNCCert(pk, prkey, sig);
		P2PWNCKey prov = new P2PWNCKey(P2PWNC_ALGO_RSA, 1024, "provider".length());
		prov.setData("provider".getBytes());
		P2PWNCReceipt rcpt = new P2PWNCReceipt(cert, prov, (int)(d.getTime()/1000), 1000, sig);

		P2PWNCConverter c = new P2PWNCConverter();
		String text = c.receiptToTextMsg(rcpt);
		System.out.println(text);


		P2PWNCReceipt r2 = c.textMsgToReceipt(text);
		if (r2 != null) {
			text = c.receiptToTextMsg(r2);
			System.out.println(text);
		}
	}
}
